@@ -11,6 +11,8 @@ import numpy as np
 import pandas as pd
 import os 
 import re
+import io
+
 
 #%% function to compute geometric factor - Jamyd91
 def geom_fac(C1,C2,P1,P2):
@@ -99,7 +101,7 @@ def geom_factor_3D(df, elec, array_type):
 
 #%% usual syscal parser
 def syscalParser(fname):#, spacing=None):
-        df = pd.read_csv(fname, skipinitialspace=True)
+        df = pd.read_csv(fname, skipinitialspace=True, engine='python', encoding_errors='ignore')
         # delete space at the end and the beginning of columns names
         headers = df.columns
         if 'Spa.1' in headers:
@@ -115,7 +117,7 @@ def syscalParser(fname):#, spacing=None):
                                     'Dev.':'dev',
                                     'M':'ip', #M1, M2,...Mn are good for now when importing
                                     'Sp':'sp'})
-        else:
+        elif 'xA(m)' in headers: # latest Prosys II
             newheaders = list(map(str.strip, headers)) 
             dico = dict(zip(headers, newheaders))
             df = df.rename(index=str, columns=dico)
@@ -128,6 +130,57 @@ def syscalParser(fname):#, spacing=None):
                                     'SP (mV)':'sp',
                                     'VMN (mV)':'vp',
                                     'IAB (mA)':'i',
+                                    'yA (m)':'ya', # new Syscal format supports topography and 3D output in the csv file
+                                    'yB (m)':'yb',
+                                    'yM (m)':'ym',
+                                    'yN (m)':'yn',
+                                    'zA (m)':'za',
+                                    'zB (m)':'zb',
+                                    'zM (m)':'zm',
+                                    'zN (m)':'zn',
+                                    'M1 (mV/V)':'M1',
+                                    'M2 (mV/V)':'M2',
+                                    'M3 (mV/V)':'M3',
+                                    'M4 (mV/V)':'M4',
+                                    'M5 (mV/V)':'M5',
+                                    'M6 (mV/V)':'M6',
+                                    'M7 (mV/V)':'M7',
+                                    'M8 (mV/V)':'M8',
+                                    'M9 (mV/V)':'M9',
+                                    'M10 (mV/V)':'M10',
+                                    'M11 (mV/V)':'M11',
+                                    'M12 (mV/V)':'M12',
+                                    'M13 (mV/V)':'M13',
+                                    'M14 (mV/V)':'M14',
+                                    'M15 (mV/V)':'M15',
+                                    'M16 (mV/V)':'M16',
+                                    'M17 (mV/V)':'M17',
+                                    'M18 (mV/V)':'M18',
+                                    'M19 (mV/V)':'M19',
+                                    'M20 (mV/V)':'M20',
+                                    'TM1 (ms)':'TM1'})
+        
+        else: # Prosys III format
+            newheaders = list(map(str.strip, headers)) 
+            dico = dict(zip(headers, newheaders))
+            df = df.rename(index=str, columns=dico)
+            df = df.rename(columns={'xA (m)':'a',
+                                    'xB (m)':'b',
+                                    'xM (m)':'m',
+                                    'xN (m)':'n',
+                                    'Dev. Rho (%)':'dev', # there is also Dev. M
+                                    'M (mV/V)':'ip',
+                                    'SP (mV)':'sp',
+                                    'VMN (mV)':'vp',
+                                    'IAB (mA)':'i',
+                                    'yA (m)':'ya', # new Syscal format supports topography and 3D output in the csv file
+                                    'yB (m)':'yb',
+                                    'yM (m)':'ym',
+                                    'yN (m)':'yn',
+                                    'zA (m)':'za',
+                                    'zB (m)':'zb',
+                                    'zM (m)':'zm',
+                                    'zN (m)':'zn',
                                     'M1 (mV/V)':'M1',
                                     'M2 (mV/V)':'M2',
                                     'M3 (mV/V)':'M3',
@@ -151,26 +204,71 @@ def syscalParser(fname):#, spacing=None):
                                     'TM1 (ms)':'TM1'})
     
         df['resist'] = df['vp']/df['i']
-
-        # if spacing is None:    
-        # for unregularly spaced array
-        array = df[['a','b','m','n']].values
         
-        # get unique electrode positions and create ordered labels for them
-        val = np.sort(np.unique(array.flatten()))
-        elecLabel = 1 + np.arange(len(val))
-        newval = elecLabel[np.searchsorted(val, array)] # magic ! https://stackoverflow.com/questions/47171356/replace-values-in-numpy-array-based-on-dictionary-and-avoid-overlap-between-new
-        df.loc[:,['a','b','m','n']] = newval # assign new label
+        # find if input contains 3D coordinates
+        if 'ya' in df.columns: # it's a 3D format file - not necessary a 3D survey!
+            syscal3D = True
+        else: # it's a 2D format file
+            syscal3D = False
         
-        # build electrode array
-        elec = np.c_[val, np.zeros((len(val), 2))]
+        if syscal3D is False:
+            # if spacing is None:    
+            # for unregularly spaced array
+            array = df[['a','b','m','n']].values
+            
+            # get unique electrode positions and create ordered labels for them
+            val = np.sort(np.unique(array.flatten()))
+            elecLabel = 1 + np.arange(len(val))
+            searchsoterdArr = np.searchsorted(val, array)
+            newval = elecLabel[searchsoterdArr] # magic ! https://stackoverflow.com/questions/47171356/replace-values-in-numpy-array-based-on-dictionary-and-avoid-overlap-between-new
+            df.loc[:,['a','b','m','n']] = newval # assign new label
+            
+            # build electrode array
+            if 'za' in df.columns and not np.all(df[['za','zb','zm','zn']].values == 0): # see if we have topography
+                zarray = df[['za','zb','zm','zn']].values
+                zvalflat = np.c_[searchsoterdArr.flatten(), zarray.flatten()]
+                zval = np.unique(zvalflat[zvalflat[:,0].argsort()], axis=0)[:,1]
+            else:
+                zval = np.zeros_like(val)
+                
+            yval = np.zeros_like(val) # 2D so Y values are all zeros
+            elec = np.c_[val, yval, zval]
+        
+        else: # we have 3D format file
+            df = df.rename(columns={'a':'xa','b':'xb','m':'xm','n':'xn'})
+            xarray = df[['xa','xb','xm','xn']].values.flatten()
+            yarray = df[['ya','yb','ym','yn']].values.flatten()
+            zarray = df[['za','zb','zm','zn']].values.flatten()
+            arrayFull = np.c_[xarray, yarray, zarray]
+    
+            elec = np.unique(arrayFull, axis=0)
+    
+            elecdf = pd.DataFrame(np.c_[elec, 1 + np.arange(len(elec))], columns=['x', 'y', 'z', 'elecnum'])
+            elecs = ['a','b','m','n']
+            for i in range(4):
+                df = df.merge(elecdf, how='outer', left_on=['x%s' % elecs[i],'y%s' % elecs[i],'z%s' % elecs[i]], right_on=['x', 'y', 'z'])
+                df = df.rename(columns={'elecnum':'%s' % elecs[i]})
+                df = df.drop(['x', 'y', 'z'], axis=1)
+                
+            df = df.dropna(subset=['a','b','m','n'])
+            df = df.astype({'a':int, 'b':int, 'm':int, 'n':int}).copy()
+            val = np.sort(elec[:,0])
+            
         # NOTE: remote electrode identification is done in R2.setElec()
         # but we notice that setting same number for remote (-99999) makes
         # the pseudo-section of remote electrode survey nicer...
-        remoteFlags = np.array([-9999999, -999999, -99999,-9999,-999,
-                                9999999, 999999, 99999, 9999, 999])
-        iremote = np.in1d(val, remoteFlags)
-        elec[iremote, 0] = -99999
+        remoteFlags_p1 = np.array([-9999999, -999999, -99999,-9999,-999])
+        remoteFlags_p2 = np.array([9999999, 999999, 99999, 9999, 999])
+        iremote_p1 = np.in1d(val, remoteFlags_p1)
+        elec[iremote_p1, 0] = -99999
+        iremote_p2 = np.in1d(val, remoteFlags_p2)
+        elec[iremote_p2, 0] = 99999
+        
+        if np.all(df['dev'].values == 0) and 'Dev. M' in df.columns: # Prosys III has to Dev. lists, assuming Dev. Rho (%) == 0 and Dev. M != 0 when IP data is collected.
+            df['dev'] = df['Dev. M'].values
+        
+        if 'ip' not in df.columns: # incase they forgot to enable 'M (mV/V)' column in Prosys
+            df['ip'] = np.nan
         
         # val = np.sort(np.unique(array.flatten()))
         # iremote = np.in1d(val, remoteFlags)        
@@ -248,8 +346,6 @@ def protocolParserLME(fname): # pragma: no cover
 # test code
 #protocolParserLME('api/test/protocol-lmeOut.dat')
 
-
-
 #%% protocol parser for 2D/3D and DC/IP
 def protocolParser(fname, ip=False, fwd=False):
     """
@@ -266,8 +362,47 @@ def protocolParser(fname, ip=False, fwd=False):
     IP 3D         11
     IP 3D + err   13
     IP 3D + fwd   12
+    
+    format:
+    R2   :5,7,7,7,7,20,15
+    cR2  :4,4,4,4,4,16,14,16
+    R3t  :5,7,4,7,4,7,4,7,4,20,15
+    cR3t :5,7,4,7,4,7,4,7,4,20,15,15
     """
-    x = np.genfromtxt(fname, skip_header=1) # we don't know if it's tab or white-space
+    # method 1: np.genfromtxt and fallback to pd.read_fwf
+    try:
+        # this should work in most cases when there is no large numbers
+        # that mask the space between columns
+        x = np.genfromtxt(fname, skip_header=1) # we don't know if it's tab or white-space
+    except Exception as e: # if no space between columns (because of big numbers or so, we fall back)
+        # more robust but not flexible to other format, this case should only be met in fwd
+        # we hope to be able to determine ncols from genfromtxt()
+        a = np.genfromtxt(fname, skip_header=1, max_rows=2)
+        threed = a.shape[1] >= 10
+        if threed is False and ip is False: # 2D DC
+            x = pd.read_fwf(fname, skiprows=1, header=None, widths=[5,7,7,7,7,20,15]).values
+        elif threed is False and ip is True: # 2D IP
+            x = pd.read_fwf(fname, skiprows=1, header=None, widths=[4,4,4,4,4,16,14,16]).values
+        elif threed is True and ip is False: # 3D DC
+            x = pd.read_fwf(fname, skiprows=1, header=None, widths=[5,7,4,7,4,7,4,7,4,20,15]).values
+        elif threed is True and ip is True: # 3D IP
+            x = pd.read_fwf(fname, skiprows=1, header=None, widths=[5,7,4,7,4,7,4,7,4,20,15,15]).values
+        else:
+            raise ValueError('protocolParser Error:', e)
+    
+    # method 2: regex (more flexible but cannot fallback to fwf)
+    # below gets all numbers possible regardless of spacing or delimiter
+    # with open(fname, "r") as f:
+    #     dump = f.readlines()
+    # nline = len(dump)
+    # numStr = r'[-+]?\d*\.\d*[eE]?[-+]?\d+|\d+' # all posible numbering formats
+    # numMeas = int(re.findall(numStr, dump[0])[0]) # number of measurements
+    # data_list = []
+    # for val in dump[1:numMeas+1]: # reding data
+    #     vals = re.findall(numStr, val)
+    #     data_list.append(vals)
+    # x = np.array(data_list).astype(float) # getting the data array
+    
     if len(x.shape) == 1: # a single quadrupole
         x = x[None,:]
     if fwd:
@@ -361,6 +496,18 @@ def primeParserTab(fname, espacing=1):
     #Note R2 expects the electrode format in the form:
     #meas.no | P+ | P- | C+ | C- | transfer resistance
     
+    def f_fill(df): # forward fill fix  
+        force_type ={'pt_p1_no:':'int64',
+                     'pt_p2_no:':'int64',
+                     'pt_c1_no:':'int64',
+                     'pt_c2_no:':'int64',
+                     'pt_calc_res:':'float64',
+                     'pt_meas_contact_resistance:':'float64'} 
+        df[['pt_c1_no:', 'pt_c2_no:']] = df[['pt_c1_no:', 'pt_c2_no:']].fillna(method='ffill', axis=0) 
+        return df.astype(force_type)
+    # fill in empty values in current electrode columns 
+    temp = f_fill(temp)  
+    
     a = temp["pt_p1_no:"]
     b = temp["pt_p2_no:"]
     m = temp["pt_c1_no:"]
@@ -372,10 +519,11 @@ def primeParserTab(fname, espacing=1):
     data_dict['dev'] = temp["pt_calc_res_error:"]
     data_dict["Rho"] = [float("nan")]*num_meas
     data_dict["ip"] = [0]*num_meas
+    data_dict['cR'] = temp['pt_meas_contact_resistance:']# contact resistance 
     df = pd.DataFrame(data=data_dict) # make a data frame from dictionary
-    df = df[['a','b','m','n','Rho','dev','ip','resist']] # reorder columns to be consistent with the syscal parser
+    df = df[['a','b','m','n','Rho','dev','ip','resist','cR']] # reorder columns to be consistent with the syscal parser
     
-    # shift electrodes to have continuous numbering
+    # shift electrodes to have continuous numbering? (not active)
     # array = df[['a','b','m','n']].values
     # arrayMin = np.min(np.unique(np.sort(array.flatten())))
     # if arrayMin != 0: # all surveys must start from x = 0
@@ -1120,7 +1268,7 @@ def stingParser(fname):
         m = 0
         while 2*m + 1 <= len(elecdf_lines) - 1: # electrodes are laid out like a snake - although not sure if this is correct
             i = 2*m + 1 # index of odd lines
-            elecdf_lines[i].loc[:, 'x'] = elecdf_lines[i].loc[:, 'x'].values[::-1].copy()
+            elecdf_lines[i].loc[:, 'x'] = elecdf_lines[i]['x'].values[::-1].copy()
             m += 1
         ######### NOT SURE ABOUT ABOVE #########
         
@@ -1195,7 +1343,7 @@ def stingParser(fname):
     #detecting IP (col 21 would be floats of IP vals)
     try:
         float(df_raw.iloc[0,21])
-        df['ip']=df_raw.iloc[0,21]
+        df['ip']=df_raw.iloc[:,21]
     except:
         df['ip']=[0]*len(df_raw)
     
@@ -1447,8 +1595,8 @@ def lippmannParser(fname):
     elec_lineNum_s = [i for i in range(len(dump)) if '* Electrode positions *' in dump[i]]
     elec_lineNum_e = [i-1 for i in range(len(dump)) if '* Remote electrode positions *' in dump[i]]
     elec_nrows = elec_lineNum_e[0] - elec_lineNum_s[0]
-    elec_raw = pd.read_csv(fname, delim_whitespace=True, skiprows=elec_lineNum_s[0]+1, nrows=elec_nrows, header=None)
-    elec = np.array(elec_raw.iloc[:,-3:])
+    elec_raw = pd.read_csv(fname, skiprows=elec_lineNum_s[0]+1, nrows=elec_nrows, header=None)[0].str.split('=', expand=True)[1]
+    elec = np.array(elec_raw.str.split(expand=True).astype(float))
     
     #for pole-pole and pole-dipole arrays
     elec[elec > 9999] = 999999
@@ -1491,11 +1639,13 @@ def aresParser(fname, spacing=None):
     #getting spacing
     spacing_lineNum = [i for i in range(len(dump)) if 'Electrode distance' in dump[i]]
     if spacing_lineNum != []:
-        spacing = dump[spacing_lineNum[0]].split()[2]
-
+        spacing = dump[spacing_lineNum[0]].split()[2] 
+    
     #getting data
     data_linNum_s = [i for i in range(len(dump)) if 'Measured data' in dump[i]]
-    df = pd.read_csv(fname, delim_whitespace=True, skiprows=data_linNum_s[0]+1, index_col=False)
+    dump_clean = [x.split() for x in dump[data_linNum_s[0]+1:] if 'too low current' not in x]  # filtering "too low current" values
+    # df = pd.read_csv(fname, delim_whitespace=True, skiprows=data_linNum_s[0]+1, index_col=False) # old way
+    df = pd.DataFrame(dump_clean[1:], columns=dump_clean[0])
     df = df.rename(columns={'C1[el]':'a',
                             'C2[el]':'b',
                             'P1[el]':'m',
@@ -1504,31 +1654,15 @@ def aresParser(fname, spacing=None):
                             'U[mV]':'vp'})
     
     #Building electrode locations
-    df[['a','b','m','n']] = df[['a','b','m','n']].apply(pd.to_numeric, errors='coerce') # there are strange eletrode conbinations sometimes (e.g., 14*1)
-    df = df.dropna()
-    array = df[['a','b','m','n']].values
-    
-    # arrayMin = np.min(np.unique(np.sort(array.flatten())))
-    # if arrayMin != 0: # all surveys must start from x = 0
-    #     array -= arrayMin
-    val = np.sort(np.unique(array.flatten())) # unique electrodes positions/labels required are 
-    # elecLabel = 1 + np.arange(len(val))
-    # newval = elecLabel[np.searchsorted(val, array)] # magic ! https://stackoverflow.com/questions/47171356/replace-values-in-numpy-array-based-on-dictionary-and-avoid-overlap-between-new
-    # df.loc[:,['a','b','m','n']] = newval
-    
-    #calculations
-#    if 'EP[mV]' in df.columns: # TODO: correct this when you figure out how the IP values are calculated EP may acctualy be SP!!
-#        df['ip'] = df['EP[mV]']/df['vp'] # not sure this is correct way of calculating IP
-#        df = df[['a','b','m','n','i','vp','ip']]
-#    else:
-    
+    df[['a','b','m','n']] = df[['a','b','m','n']].replace('\*\d+', '', regex=True).astype(int) # there are multi eletrode conbinations sometimes (e.g., 14*1) which basically mean the middle one
+    df = df.dropna(subset=['a', 'b', 'm', 'n', 'Pn', 'Pn+1', 'Array', 'Uout[V]', 'i', 'vp', 'EP[mV]', 'AppRes[Ohmm]', 'St-dev[%]']) # some IP cols are None anyways per sequence type
+
     
     # finding IP columns
-    
     ip_cols = [col for col in df.columns if all(i in col for i in ['IP', '[%]'])] # assuming IP columns are those with IP and [%]
     if ip_cols!= []:
         # df[ip_cols] = df[ip_cols]/100 # not sure about this... who reports IP in % anyway?!
-        df['ip'] = df[ip_cols].mean(axis=1) # average of all IP windows
+        df['ip'] = df[ip_cols].astype(float).mean(axis=1) # average of all IP windows
         df = df[['a','b','m','n','i','vp','ip']]
         
     else:
@@ -1536,10 +1670,20 @@ def aresParser(fname, spacing=None):
         df['ip'] = 0 # should be under "else:"
     
     df = df.query("i != '-' & vp != '-' & ip != '-'").astype(float)    
+    df[['a','b','m','n']] = df[['a','b','m','n']].astype(int)
     
     df['resist'] = df['vp']/df['i']
     
-    #builting electrode table
+    # df = df.dropna()
+    
+    #building electrode table
+    array = df[['a','b','m','n']].values
+    val = np.sort(np.unique(array.flatten()))
+    elecLabel = 1 + np.arange(len(val))
+    searchsoterdArr = np.searchsorted(val, array)
+    newval = elecLabel[searchsoterdArr] # magic ! https://stackoverflow.com/questions/47171356/replace-values-in-numpy-array-based-on-dictionary-and-avoid-overlap-between-new
+    df.loc[:,['a','b','m','n']] = newval # assign new label
+    
     if spacing is not None:
         elec = np.c_[val*float(spacing), np.zeros((len(val),2))]
     else:
@@ -1557,6 +1701,7 @@ def bertParser(fname):
     f = open(fname, "r")
     
     dump = f.readlines()
+    f.close()
     line = 0
     
     # skip comment lines
@@ -1593,7 +1738,7 @@ def bertParser(fname):
         data_input_raw = dump[line].split('#')[0] # getting rid of comments 
         vals = re.findall(numStr, data_input_raw)
     
-    headers = re.findall(r'[A-Za-z]+', dump[line-1]) # for finding data types
+    headers = re.findall(r'[A-Za-z\/]+', dump[line-1]) # for finding data types
 
     topo_check_vals = len(re.findall(numStr, dump[line].split('#')[0])) # TODO: is topography included without any flags?
     df_list = []
@@ -1705,5 +1850,155 @@ def srvParser(fname):
     data_dict['ip']=[0]*nmeas
     df = pd.DataFrame(data=data_dict) # make a data frame from dictionary
     df = df[['a','b','m','n','Rho','dev','ip','resist','magErr']] # reorder columns to be consistent with the syscal parser
+    fh.close()
     
     return elec, df 
+
+#%% DAS-1 parser
+def dasParser(fname):
+    with open(fname, "r") as f:
+        dump_raw = f.readlines()
+
+    numStr = r'[-+]?\d*\.\d*[eE]?[-+]?\d+|\d+' # all posible numbering formats
+    
+    # cleaning data from "out of range" measurements
+    dump = [val for val in dump_raw if 'out of range' not in val]
+    cleanData = ''.join(dump) # remove lines without data
+    
+    # getting electrode locations and creating dfelec
+    elec_lineNum_s = [i+2 for i in range(len(dump)) if '#elec_start' in dump[i]] # assuming headers !Cbl# El# Elec-X  Elec-Y  Elec-Z  Terrn-Z  Type  El.Num
+    elec_lineNum_e = [i for i in range(len(dump)) if '#elec_end' in dump[i]]
+    nrows = elec_lineNum_e[0] - elec_lineNum_s[0]
+    dfElec_raw = pd.read_csv(io.StringIO(cleanData), delim_whitespace=True, skiprows=elec_lineNum_s[0], 
+                             nrows=nrows, index_col=False, header=None)
+    elecNum = dfElec_raw.iloc[:,0].str.split(',', expand=True).astype(int).astype(str)
+    elecLabel = elecNum[0].str.cat(elecNum[1], sep=' ')
+    dfelec = pd.DataFrame()
+    dfelec['label'] = elecLabel.copy()
+    dfelec[['x', 'y', 'z']] = dfElec_raw.iloc[:,1:4]
+    dfelec['buried'] = False
+    dfelec['remote'] = False
+    
+    # remote electrodes?
+    remote_flags = [-9999999, -999999, -99999,-9999,-999,
+                9999999, 999999, 99999] # values asssociated with remote electrodes
+    iremote = np.in1d(dfelec['x'].values, remote_flags)
+    iremote = np.isinf(dfelec[['x','y','z']].values).any(1) | iremote
+    dfelec.loc[:, 'remote'] = iremote
+    
+    # getting data and creating df
+    df_lineNum_s = [i+3 for i in range(len(dump)) if '#data_start' in dump[i]] # assuming headers exist
+    df_lineNum_e = [i for i in range(len(dump)) if '#data_end' in dump[i]]
+    
+    df_list = []
+    for val in dump[df_lineNum_s[0]:df_lineNum_e[0]]: # reding data from mixed separated numbers!
+        vals = re.findall(numStr, val)
+        df_list.append(vals)
+    df_raw = pd.DataFrame(np.array(df_list).astype(float))
+    
+    # 2D or 3D? 2D will have only one line in line numbers for A, B, M or N
+    flagD = '3D' if np.mean(df_raw.iloc[:,1]) != df_raw.iloc[0,1] else '2D'
+    
+    # find data cols from file headers
+    resCol = int([val.split()[-1] for val in dump if 'data_res_col' in val][0]) - 1 # -1 for python numbering
+    devCol = int([val.split()[-1] for val in dump if 'data_std_res_col' in val][0]) - 1 
+    ipCol = int([val.split()[-1] for val in dump if 'data_ip_wind_col' in val][0]) - 1 # this may become problematic if no IP in data
+        
+    df = pd.DataFrame()
+    arrHeader = ['a', 'b', 'm', 'n']
+    
+    #2D array
+    if flagD == '2D':
+        lineNumber = int(df_raw.iloc[0,1]) # DAS-1 can have random line numners and must mach with available electrodes
+        elecLNum = elecNum[0].astype(int)
+        selectElecs = elecLNum[elecLNum == lineNumber].index.values # selected elec indecis 
+        dfelec = dfelec.iloc[selectElecs,:].reset_index(drop=True)
+        
+        # may need convert 3D XYZ into 2D!!
+        if np.mean(dfelec['x']) == dfelec['x'][0]:
+            dfelec['x'] = dfelec['y'].values.copy()
+            dfelec['y'] = 0
+        
+        # sorting needed
+        dfelec = dfelec.sort_values('x').reset_index(drop=True)
+        dfelec = dfelec[['x', 'y', 'z']].values # it's easier for 2D to skip labels, etc.
+        
+        # array
+        for x in enumerate(arrHeader):
+            df[x[1]] = df_raw.iloc[:,(x[0]+1)*2].astype(int).copy()
+        
+    #3D array
+    elif flagD == '3D':
+        # getting correct lines from elec table!
+        lines = np.unique(df_raw.iloc[:,1].values)
+        elecLNum = elecNum[0].astype(int)
+        dfelec_selected = []
+        for lineNumber in lines:
+            selectElecs = elecLNum[elecLNum == lineNumber].index.values # selected elec indecis 
+            dfelec_selected.append(dfelec.iloc[selectElecs,:])
+        
+        dfelec = pd.concat(dfelec_selected).reset_index(drop=True) # recreating dfelec
+            
+        for x in enumerate(arrHeader):
+            left = df_raw.iloc[:,(x[0]*2)+1].astype(int).astype(str) # creating protocol-like labels
+            right = df_raw.iloc[:,(x[0]+1)*2].astype(int).astype(str)
+            df[x[1]] = left.str.cat(right, sep=' ')
+        
+    # data
+    df['resist'] = df_raw[resCol].values
+    if devCol >= 0:
+        df['dev'] = df_raw[devCol].values
+    df['ip'] = df_raw[ipCol].values if ipCol > 1 else 0
+    
+    return dfelec, df
+    
+    
+def geomParser(fname):
+    """
+    Parse PRIME Geometry file. 
+
+    Parameters
+    ----------
+    fname : str
+        Filepath to file.
+
+    Returns
+    -------
+    dfelec: pd.Dataframe() 
+        Geometry dataframe 
+    """
+    fh = open(fname,'r')
+    
+    x = []
+    y = []
+    z = []
+    n = []
+    b = []
+    
+    c = 0 
+    line = fh.readline()
+    # find header line 
+    while 'data:' not in line: 
+        line = fh.readline() 
+        c+=1 
+        if c>100:
+            raise Exception('Cannot find data line in geom file')
+            break 
+    _ = fh.readline()
+    # read in geometry information 
+    info = fh.readline().split() 
+    while len(info)>0:
+        n.append(int(info[0]))
+        x.append(float(info[1]))
+        y.append(float(info[2]))
+        z.append(float(info[3]))
+        if info[-1] =='Surface': 
+            b.append(0)
+        else:
+            b.append(1)
+        info = fh.readline().split()
+    # convert to pandas dataframe format 
+    d = {'x':x,'y':y,'z':z,'buried':b}
+    dfelec = pd.DataFrame(d)
+    return dfelec 
+    
\ No newline at end of file
